from print_ext.fill import Fill
from print_ext.line import SMark as SM

def _f(*args, w=0, h=0, **kwargs):
    return ','.join(map(str, Fill(*args, **kwargs).flatten(w,h)))

def test_fill_unicode():
    assert(_f('あかさ', w=8,h=3) == 'あかさあ,あかさあ,あかさあ')
    assert(_f('か', w=1,h=3) == ' , , ')


def test_fill_different_line_width():
    assert(_f('あかa\vxy', w=3, h=3) == 'あ ,xyx,あ ')
    assert(_f('xy\v', w=3, h=3) == 'xyx,   ,xyx')
    assert(_f('abcd','\ve', w=0, h=4) == 'abcd,eeee,abcd,eeee')
    assert(_f('abcd\ve', w=0, h=0) == 'abcd,eeee')


def test_fills():
    assert(_f('', w=3, h=3) == '   ,   ,   ')
    assert(_f('abcde', w=3, h=2) == 'abc,abc')
    assert(_f(w=3, h=2) == '   ,   ')


def test_fill_style():
    f = Fill('\b2 xy', 'z', style='1')
    assert(list(f.flatten(w=7))[0].styled() == ('xyzxyzx', [SM('1',0,7), SM('2',0,2), SM('2',3,5), SM('2',6,7)]))


def test_fill_doctest():
    assert([str(l) for l in Fill().flatten(2,2)] == ['  ', '  '])
    assert([str(l) for l in Fill('abc').flatten(5,2)] == ['abcab','abcab'])
    assert([str(l) for l in Fill('a\v','b').flatten(3,3)] == ['aaa','bbb','aaa'])



import pytest, io
from print_ext.printer import Printer, stack_enum
from print_ext.line import SMark as SM


class PrinterTest(Printer):
    def format_out(self, txt, styles):
        print(f'format_out "{txt}"  styles:{styles}')
        stripped = txt.rstrip()
        s = ''
        for t,stk in stack_enum(txt, styles):
            s += stripped[:len(t)] + ('_' if not stk else ''.join(stk))
            stripped = stripped[len(t):]
        return s



def _test_printer(**kwargs):
    o = io.StringIO()
    p = PrinterTest(stream=o, **kwargs)
    return o,p



def _printer(**kwargs):
    o = io.StringIO()
    p = Printer(stream=o, **kwargs)
    return o,p
   


def test_stack_enum():
    def _tst(*args):
        sout = ''
        for s,stk in stack_enum(*args):
            sout += f"{s}[{''.join(stk)}]"
        return sout
    # This is impossible from styled() assert(_tst('abcdef', [SM('0',0,3), SM('0',3,6)]) == 'abcdef[0]')
    assert(_tst('abcdef', [SM('0',0,6), SM('1',0,6),SM('',1,5),SM('',2,4)]) == 'a[01]b[0]cd[]e[0]f[01]')
    assert(_tst('abcdef', [SM('1',1,4)]) == 'a[]bcd[1]ef[]')
    assert(_tst('abcdef', [SM('0',0,1), SM('1',1,4),SM('2',1,3)]) == 'a[0]bc[12]d[1]ef[]')
    assert(_tst('abcdef', [SM('0',0,1), SM('1',1,4),SM('2',1,3)]) == 'a[0]bc[12]d[1]ef[]')
    assert(_tst('abcdef', [SM('0',0,6), SM('1',3,6)]) == 'abc[0]def[01]')
    assert(_tst('abcdef', [SM('0',0,6), SM('1',2,5)]) == 'ab[0]cde[01]f[0]')
    assert(_tst('abcdef', [SM('0',0,6), SM('1',0,6)]) == 'abcdef[01]')



def test_stream_out():
    o,p = _test_printer()
    p.stream_out([33, 44, 'bob    ', ' cob', ''])
    assert(repr(o.getvalue()) == r"'33\n44\nbob\n cob\n\n'")
    p.stream_out(['x\n\n','',''])
    assert(repr(o.getvalue()) == r"'33\n44\nbob\n cob\n\nx\n\n\n'")
    p.stream_out(['','','y',''])
    assert(repr(o.getvalue()) == r"'33\n44\nbob\n cob\n\nx\n\n\ny\n\n'")
    p.stream_out(['','','z'])
    assert(repr(o.getvalue()) == r"'33\n44\nbob\n cob\n\nx\n\n\ny\n\n\nz\n'")



def test_print_call():
    o,p = _test_printer()
    p('x','y','\b3 z     ')
    assert(o.getvalue() == 'xy_z3\n')
    o,p = _test_printer()
    p(''' Multi
    Line        
    Comment''')
    assert(repr(o.getvalue()) == r"'Multi_\nLine_\nComment_\n'")


@pytest.mark.skip(reason="Not implemented")
def test_lang():
    o,p = _test_printer(lang='JA')
    p('hello\fja こんにちは')
    assert(repr(o.getvalue()) == r"'こんにちは_\n'")


def test_color():
    o,p = _test_printer(lang='JA')
    p('\berr ERROR',' j/k')
    assert(repr(o.getvalue()) == r"'ERRORerr j/k_\n'")
    

def test_printer_default_styles():
    o,p = _printer(color=True)
    p('the ', '\bem-_ Quick\vbrown', ' fox')
    print(o.getvalue())
    assert(repr(o.getvalue()) == repr('the \x1b[1;4mQuick\x1b[0m\n\x1b[1;4mbrown\x1b[0m fox\n'))
    o,p = _printer()
    p('the ', '\br_ Quick\vbrown', ' fox')
    print(o.getvalue())
    assert(repr(o.getvalue()) == repr('the Quick\nbrown fox\n'))


def test_printer_default_underscore():
    o,p = _printer(color=True)
    p('the \br_$ quick ','\b!\b;$ brown', ' fox')
    print(o.getvalue())
    assert(repr(o.getvalue()) == repr('the \x1b[4;31mquick \x1b[1;2mbrown\x1b[0;2;4;31m fox\x1b[0m\n'))
    

def test_printer_default_bold():
    o,p = _printer(color=True)
    p('the ', '\b! quick',' fox')
    print(o.getvalue())
    assert(repr(o.getvalue()) == repr('the \x1b[1mquick\x1b[0m fox\n'))
    

def test_printer_to_str():
    from print_ext import print
    assert(print.to_str('hello', ' \b1 world') == 'hello world\n')



from .rich import Rich
from .flex import Flex
from .line import Line, StyleCVar
from .context import Context, CVar
from .border_dfn import BorderDfn


class BorderCVar(CVar):
    def canon(self, val):
        if isinstance(val, tuple): return BorderDfn(*val)
        if isinstance(val, dict): return BorderDfn(**val)
        return BorderDfn(val)

    def merge(self, val, pval):
        return pval.update(val)


Context.define(StyleCVar('border_style'))
Context.define(BorderCVar('border'))



class Borders(Flex, border=(1,'-'), border_style='dem'):

    def calc_width(self):
        bdr = self['border']
        return super().calc_width(Flex) + (bdr.width if bdr else 0)

    
    def calc_height(self):
        bdr = self['border']
        return super().calc_height(Flex) + (bdr.height if bdr else 0)


    def flatten(self, w=0, h=0, **kwargs):
        bdr = self['border']
        bdrw = w==0 or w >= bdr.width + 1
        bdrh = h==0 or h >= bdr.height + 1
        bw = int(bdrw and bdr.width)
        bh = int(bdrh and bdr.height)
        child_size = (w and w-bw), (h and h-bh)
        flat = super().flatten(w=child_size[0], h=child_size[1], **kwargs)# if self.child else []
        ascii = self['ascii']
        style = self['border_style']
        lhs = bdr.side_chars(len(flat), bdr.l[5:] if ascii else bdr.l) if bdr.sides[2] else ''
        rhs = bdr.side_chars(len(flat), bdr.r[5:] if ascii else bdr.r) if bdr.sides[3] else ''
        wb = w or (flat[0].width if flat else 0) + int(bdr.sides[2]) + int(bdr.sides[3])
        #print(f'borders flatten {w}x{h}  bdwh:{bw}x{bh}  {child_size}  ->  bw:{wb} {lhs!r} {rhs!r}')
        if bdrh and (txt:=bdr.top_line(wb, ascii)):
            yield Line(parent=self, style=style).insert(0, txt)
        if not lhs and not rhs:
            yield from flat
        else:
            for i, f in enumerate(flat):
                l = Line(parent=self)
                if lhs: l(lhs[i], style=style)
                l(f)
                if rhs: l(rhs[i], style=style)
                yield l
        if bdrh and (txt:=bdr.bottom_line(wb, ascii)):
            yield Line(parent=self, style=style).insert(0, txt)



import os, sys, locale, io
from functools import reduce
from .borders import Borders
from .table import Table
from .context import Context
from .flex import Flex
from .text import Text
from .pretty import pretty
from .sgr import SGR
from .hr import HR
from .card import Card
from .progress import Progress
from .widget import INFINITY


def stack_enum(txt, stack):
    a, b = 0, 0
    _pop, stk = [], []
    while a < len(txt):
        if a != b:
            yield txt[a:b], stk
            a = b
        # Apply pop styles to stk
        while _pop and _pop[-1].e == b:
            j, s = 0, _pop.pop().style
            while not s: j, s = j+1, _pop[-j-1].style
            stk.append(_pop[-2*(j-1)-1].style) if j else stk.pop()
        # Apply pre styles to stk 
        while stack and stack[0].s == b:
            _pop.append(stack.pop(0))
            stk.append(_pop[-1].style) if _pop[-1].style else stk.pop()
        # Advance to the next style boundry
        b = min(stack[0].s if stack else INFINITY, _pop[-1].e if _pop else INFINITY)



def _stack_enum(txt, stack):
    # De-duplicate styles
    ps, pstk = '', None
    for s, stk in _stack_enum(txt,stack):
        if pstk == None or pstk == stk:
            ps += s
        else:
            yield ps, pstk
            ps = s
        pstk = stk[:]
    yield ps, pstk



class Printer(Context):
    ''' This is an object for aliasing the print function.
    '''
    default_styles = {
        'err' : 'r!,',
        'warn': 'y!,',
        'em'  : '!',
        'dem' : 'w.;',
        '1' : 'y,',
        '2' : 'm,',
        '3' : 'c,',
    }

    print = print # The legacy print() function

    def __init__(self, *, stream=None, color=None, width=None, isatty=None, styles=default_styles, **kwargs):
        self.styles = styles
        self.stream = stream or sys.stdout
        kwargs = {k:v for k,v in kwargs.items() if v != None}
        if isatty==None:
            try:    self.isatty = self.stream.isatty()
            except: self.isatty = False
        #print(f"isatty? {isatty}: {self.stream}")
        if 'lang' not in kwargs:
            kwargs['lang'] = locale.getdefaultlocale()[0]
        kwargs['lang'] = kwargs['lang'].lower()
        if width == None:
            try:    width = os.get_terminal_size().columns-1
            except: width = INFINITY
        kwargs['width_max'] = width
        if 'ascii' not in kwargs:
            kwargs['ascii'] = (locale.getdefaultlocale()[1].lower() != 'utf-8')
        self.color = self.isatty if color == None else color
        super().__init__(**kwargs)
        self.blank = 0


    def format_out(self, txt, styles):
        stripped = txt.rstrip()
        s = ''
        sgr_prev = SGR()
        for t,stk in stack_enum(txt, styles):
            sgrs = [SGR(self.styles.get(y,y)) for y in stk] or [SGR()]
            #print(f'format_out {repr(stripped[:len(t)])} {stk} -> {sgrs}')
            if self.color:
                sgr_next = reduce(lambda a,b: a+b, sgrs)
                code = sgr_next.diff(sgr_prev)
                #print(f"         {sgr_prev} -> {sgr_next}  {repr(code)}")
                s += code
                sgr_prev = sgr_next
            s += stripped[:len(t)]
            stripped = stripped[len(t):]
        if sgr_prev: s += '\033[0m'
        return s


    def stream_out(self, lines):
        new_blank = 0
        for line in lines:
            line = self.format_out(*line.styled()) if hasattr(line, 'styled') else str(line).rstrip()
            if not line and self.blank:
                self.blank -= 1
                continue
            self.blank = 0
            new_blank = 0 if line else new_blank + 1
            self.stream.write(line+'\n')
        self.blank += new_blank


    def __call__(self, *args, **kwargs):
        t = Text(*args, parent=self, **kwargs)
        self.stream_out(t.flatten(**kwargs))


    def flex(self, *args, **kwargs):
        f = Flex(*args, parent=self, **kwargs)
        self.stream_out(f.flatten(**kwargs))


    def progress(self, *args, **kwargs):
        return Progress(*args, parent=self, **kwargs)
        

    def card(self, *args, **kwargs):
        c = Card(*args, parent=self, **kwargs)
        self.stream_out(c.flatten(**kwargs))


    def hr(self, *args, **kwargs):
        h = HR(*args, parent=self, **kwargs)
        self.stream_out(h.flatten(**kwargs))


    def pretty(self, *args, **kwargs):
        self(*(pretty(a,**kwargs) for a in args), **kwargs)


    def to_str(self, *args, **kwargs):
        saved = self.stream
        self.stream = io.StringIO()
        self(*args, **kwargs)
        val = self.stream.getvalue()
        self.stream = saved
        return val



import pytest 
from print_ext.context import Context, CVar, IntCVar, FloatCVar, ObjectAttr
from .testutil import debug_dump, print_ctx_trace

class MergeCVar(CVar):
    def canon(self, v):
        return v if isinstance(v, list) else [v]

    def merge(self, val, pval):
        return val + pval


Context.define(CVar('d'))
Context.define(CVar('z'))
Context.define(IntCVar('m'))
Context.define(IntCVar('q','Q'))
Context.define(IntCVar('a','A','aa'))
Context.define(FloatCVar('pi','PI'))
Context.define(MergeCVar('r'))


class Animal(Context, a=3, pi=3.1415, r='x'):
    pass#ctx_defaults = Context.defaults()


class Cat(Animal, A=4, r=3):
    #ctx_defaults = Context.defaults(A=4)

    @property
    def width(self):
        return 99 + self['a']


class Dog(Animal, a=5, r='hi'):
    pass#ctx_defaults = Context.defaults(a=5)


def test_ctx_redefine():
    with pytest.raises(ValueError):
        Context.define(IntCVar('A','a','aa'))
    with pytest.raises(ValueError):
        Context.define(IntCVar('a','A'))
    Context.define(IntCVar('a','aa','A'))



def test_ctx_CallableVar():
    c = Animal(width_nom=ObjectAttr('width', -101))
    assert(c['width_nom'] == -101)
    c = Cat(width_nom=ObjectAttr('width', -101))
    assert(c['width_nom'] == 103)



def test_ctx_class_merge():
    c = Cat(r='z')
    d = Dog(parent=c)
    print_ctx_trace(d)
    assert(d['r'] == ['z', 'hi','x', 3, 'x'])



def test_ctx_parent_loops():
    ''' Don't create loops
    '''
    c = Cat(pi=3.1)
    d = Dog(parent=c)
    a = Animal(parent=d, a=44, q=5)
    assert(c['a'] == 4)
    assert(c['q'] == None)
    assert(a['pi'] == 3.1) 
    idc = id(c)
    idc2 = c.ctx_parent(a)
    assert(id(idc2) != idc)
    assert(idc2['a'] == 4)
    assert(idc2['q'] == 5)
    
    cc = Cat(pi=4)
    dd = Dog(parent=cc)
    assert(dd['q'] == None)
    assert(dd['pi'] == 4)
    ddd = dd.ctx_parent(a)
    assert(ddd['q'] == 5)
    assert(ddd['pi'] == 4)



def test_kwargs():
    c = Cat(pi=3.1)
    assert(c['pi'] == 3.1)
    c['PI'] = 2
    assert(c['pi'] == 2)



def test_not_found():
    c = Cat()
    with pytest.raises(KeyError):
        c['dne']
    


def test_context_defaults():
    c = Cat()
    assert(c['a'] == 4)
    assert(c['q'] == None)
    assert(Animal()['a'] == 3)
    


def test_del():
    c = Cat(A=99)
    assert(c['a'] == 99)
    del c['a']
    assert(c['a'] == 4)


def test_parent_value():
    c1 = Cat(m=99, q=424, d='cat')
    c2 = Dog(m=33, z='zz', parent=c1)
    assert(c1['m'] == 99)
    assert(c2['m'] == 33)
    assert(c1['q'] == 424)
    assert(c2['q'] == 424)
    assert(c2['z'] == 'zz')
    assert(c2['a'] == 5)
    assert(c2['d'] == 'cat')
    del c1['d']
    assert(c2['d'] == None)


def test_ctx_empty_get():
    c1 = Cat(m=99, q=424, d='cat')
    assert(c1.ctx() == None)
    


from print_ext.context import Context
from print_ext.line import Line
from print_ext.span import Span
from print_ext.text import Text
import print_ext.flex

class Fill(Text, width_min=0, height_min=0):
    ''' Fills the given block of space (when flattened) with a simple repeating pattern.

    :Examples:

        >>> [str(l) for l in Fill().flatten(2,2)]
        ['  ', '  ']
        >>> [str(l) for l in Fill('abc').flatten(5,2)]
        ['abcab', 'abcab']
        >>> [str(l) for l in Fill('a\v','b').flatten(3,3)]
        ['aaa', 'bbb', 'aaa']
    
    more
    '''

    def calc_width(self):
        return 0


    def calc_height(self):
        return 0


    def flatten(self, w=0, h=0, **kwargs):
        rows = [x.clone(parent=self, **x.ctx_local) for x in self.lines]
        if not rows: rows = [Line(' ', parent=self)]
        if not w: w = rows[0].width if len(rows) == 1 else max(*[r.width for r in rows])
        for ri, line in enumerate(rows):
            if h and ri == h: break # slight optimization?
            if not line: line.insert(0, ' '*w)
            while line.width < w:
                line.insert(-1, line.clone(parent=line.parent, **line.ctx_local))
            line.trim(w).insert(-1, ' '*(w-line.width))
        for i in range(h or len(rows)):
            yield rows[i%len(rows)]



from print_ext.line import Line
from print_ext.context import Context
from print_ext.widget import Widget


class Bar(Widget):
    def __init__(self, chars=None, steps=1, **kwargs):
        self.i = 0
        self.steps = steps
        super().__init__(**kwargs)
        self.chars = chars or ('#-' if self['ascii'] else '━┄')
        

    def calc_width(self):
        return 35
    

    def calc_height(self):
        return 1


    def step(self, pct=None):
        self.i = (self.i+1) if pct == None else min(max(0, pct*self.steps), self.steps)


    def flatten(self, w=0, h=0, **kwargs):
        w = w or 30
        nleft = min(int(self.i * w / self.steps), w)
        return [Line(parent=self).insert(0,self.chars[0]*nleft).insert(-1,self.chars[1]*(w-nleft), style=';.')]



class Spinner(Bar):

    def __init__(self, chars=None, **kwargs):
        super().__init__(**kwargs)
        self.chars = chars or '/-\\|'
        

    def calc_width(self):
        return 3
 

    def flatten(self, w=0, h=0, **kwargs):
        i = int(self.i)%len(self.chars)
        w = w or 1
        return [Line(parent=self).insert(0,' ' + self.chars[i])]



class Progress(Context):
    def __init__(self, fn=None, steps=0, *, parent, **kwargs):
        kwargs.setdefault('style','1')
        self.fn = fn
        self.prev_width = 0
        super().__init__(parent=parent)
        self.bar = Bar(steps=steps, parent=self, **kwargs) if steps else Spinner(parent=self, **kwargs)
       

    def __call__(self, *args, done=False, style=None, **kwargs):
        self.bar.step()
        done = done or not self.parent.isatty
        #print(f'bar {self.bar!r}')
        line = Line(*args, parent=self, **kwargs)
        if not done or not isinstance(self.bar, Spinner):
            line.insert(0,' ').insert(0, self.bar, style=style).insert(0, ' ')
        txt = self.parent.format_out(*line.styled())
        pad = self.prev_width - line.width
        pad = ' '*pad if pad > 0 else ''
        self.prev_width = line.width+1
        self.parent.stream.write(txt + pad + ('\n' if done else '\r'))




import re
from math import ceil
from functools import reduce
from .flex import Flex
from .line import StyleCVar
from .context import Context, CVar
from .text import Text
from .size import Size
from .borders import Borders, BorderDfn


Context.define(CVar('cls'))
Context.define(CVar('tmpl'))


class CellDfn():
    RE = re.compile(r'(r(-?\d+)?(%(\d+))?)?(c(-?\d+)?(%(\d+))?)?$')

    def __init__(self, cstr, **kwargs):
        self.kwargs = kwargs
        cstr = cstr.lower()
        if cstr == 'all': m = (None, True, 0, True, 1, True, 0, True, 1)
        else: m = CellDfn.RE.match(cstr)
        if not m: raise ValueError(f"Invalid Cell Range {cstr!r}")
        self.r, self.rmod = (int(m[2] or 0), int(m[4]) if m[3] else 0) if m[1] else (0, 1)
        self.c, self.cmod = (int(m[6] or 0), int(m[8]) if m[7] else 0) if m[5] else (0, 1)


    def __hash__(self):
        return hash(self.vals())


    def __eq__(self, other):
        return self.vals() == (other and other.vals())


    def vals(self):
        v = self.r, self.rmod, self.c, self.cmod
        return  v + tuple((k,self.kwargs[k]) for k in sorted(self.kwargs))
    
    
    def ctx_merge(self, ctx, r, c, n_rows, n_cols):
        if not self.matches(r, c, n_rows, n_cols): return ctx
        return Context(parent=ctx, **self.kwargs)


    def matches(self, r, c, n_rows, n_cols):
        r_off = self.r % n_rows
        c_off = self.c % n_cols

        if self.rmod:
            if (r < r_off) or (r-r_off)%self.rmod: return False
        elif r_off != r: return False

        if self.cmod:
            if (c < c_off) or (c-c_off)%self.cmod: return False
        elif c_off != c: return False

        return True

    
    def is_open_ended(self):
        return self.r >= 0


    def __repr__(self):
        return f'R{self.r}%{self.rmod}C{self.c}%{self.cmod} {self.kwargs}'



class Table(Flex, tmpl='pad,em'):    
    _tmpls = {}

    @staticmethod
    def define_tmpl(style, *cells):
        if style in Table._tmpls and cells != Table._tmpls[style]:
            raise ValueError(f"Table style {style!r} is already defined:\nExisting: {Table._tmpls[style]}\nTrying to set: {cells}")
        Table._tmpls[style] = cells


    def __init__(self, *args, **kwargs):
        self.cols = [Size(easy=a) for a in args]
        self._cell_ctx = []
        super().__init__(**kwargs)


    def __bool__(self):
        return bool(self.__len__())


    def __len__(self):
        return len(self.cells)


    @property
    def cells(self):
        while self.rich_stream:
            elctx = self.rich_stream.pop(0)
            if elctx[0] == '\t':
                self._cells.append([])
            else:
                self._cells[-1].append(elctx)
        return self._cells[:-1] if self._cells[-1] == [] else self._cells



    def calc_width(self):
        n_cols = len(self.cols)
        cells = self.cells
        n_rows = ceil(len(cells)/n_cols)
        els = self._cell_instances(n_rows, n_cols)
        # Calculate max col widths
        noms = [0]*n_cols
        for i, el in enumerate(els):
            c = i%n_cols
            if noms[c] < (nom:=el['width_nom']): noms[c] = nom
        sizes = [col.clone(nom=noms[i]) for i,col in enumerate(self.cols)]
        if len(sizes) != n_cols: raise NotImplementedError()
        def elide_cols(ecells):
            return Size(nom=1, rate=0, user=ecells)
        cols = Size.resize(sizes, 0, wrap=False, elide=elide_cols)[0]
        return reduce(lambda a,c: a + c.size, cols, 0)


    def is_open_ended(self):
        return all(c.is_open_ended() for c in self._cell_ctx)


    def _cell_ctx_reduce(self, r, c, n_rows, n_cols):
        ctx = reduce(lambda a,b: b.ctx_merge(a, r, c, n_rows, n_cols), self._cell_ctx, Context())
        tmpl = [x.strip() for x in (ctx['tmpl'] or self['tmpl']).split(',') if x.strip()]
        #while '0' in tmpl: tmpl = tmpl[tmpl.index('0')+1:]
        tmpls = [a for t in tmpl for a in Table._tmpls[t]]
        if tmpls:
            ctx_pre = reduce(lambda a,b: b.ctx_merge(a, r, c, n_rows, n_cols), tmpls, Context())
            pctx = ctx
            while pctx.parent != None:
                pctx = pctx.parent
            pctx.parent = ctx_pre
        return ctx


    def _cell_instance(self, cell, ctx):
        if cls:=ctx['cls']:
            el = cls(parent=self, **ctx.ctx_flatten())
            if cell: el.rich_newline(cell[0][1]) if cell[0][0]==None else el.rich_append(*cell[0])
        elif cell and isinstance(cell[0][0], Context):
            kwargs = ctx.ctx_flatten()
            kwargs.update(cell[0][1])
            kwargs.update(cell[0][0].ctx_local)
            el = cell[0][0].clone(parent=self, **kwargs)
        else:
            el = Text(parent=self, **ctx.ctx_flatten())
            if cell: el.rich_newline(cell[0][1]) if cell[0][0]==None else el.rich_append(*cell[0])
        for args in cell[1:]:
            el.rich_newline(args[1]) if args[0]==None else el.rich_append(*args)
        return el


    def _cell_instances(self, n_rows, n_cols):
        if self.__cell_instances != None: return self.__cell_instances
        self.__cell_instances = []
        for i, cell in enumerate(self.cells):
            c = i%n_cols
            ctx = self._cell_ctx_reduce(i//n_cols, c, n_rows, n_cols)
            self.__cell_instances.append(self._cell_instance(cell, ctx))
        return self.__cell_instances
            

    def changed_size(self):
        self.__cell_instances = None
        super().changed_size()


    def flatten(self, w=0, h=0, **kwargs):
        n_cols = len(self.cols)
        cells = self.cells
        n_rows = ceil(len(cells)/n_cols)
        els = self._cell_instances(n_rows, n_cols)
        # Calculate max col widths
        noms = [0]*n_cols
        for i, el in enumerate(els):
            c = i%n_cols
            if noms[c] < (nom:=el['width_nom']): noms[c] = nom
        # Calculate sizes
        sizes = [col.clone(nom=noms[i]) for i,col in enumerate(self.cols)]
        if len(sizes) != n_cols: raise NotImplementedError()
        #print(f"table flatten {len(cells)} els into table {n_rows}x{n_cols} => {sizes}  :{w}x{h}")
        def elide_cols(ecells):
            return Size(nom=1, rate=0, user=ecells)
        cols = Size.resize(sizes, w, wrap=False, elide=elide_cols)[0]
        # Flatten
        row_els = []
        for i, e in enumerate(els):
            row_els.append(e)
            if len(row_els) == n_cols or i == len(cells)-1:
                if h:
                    raise NotImplementedError()
                else:
                    for e,c in zip(row_els + [Text(parent=self)]*(len(cols)-len(row_els)), cols): c.user = e
                    yield from self._flatten_width_row(cols, w, 0)
                row_els = []


    def cell(self, rstr, **kwargs):
        self._cell_ctx.append(CellDfn(rstr, **kwargs))


    def clone(self, **kwargs):
        s = self.__class__(**kwargs)
        s._cells = []
        for row in self.cells:
            rcells = []
            for col in row:
                if isinstance(col[0], Context):
                    rcells.append( (col[0].clone(parent=s, **col[0].ctx_local), col[1]) )
                else:
                    rcells.append( col )
            s._cells.append(rcells)


        #s._cells = [c.clone(parent=s, **c.ctx_local) for c in self.cells]
        if self._cells[-1] == []: s._cells.append([])
        s.cols = list(self.cols)
        s._cell_ctx = list(self._cell_ctx)
        return s



Table.define_tmpl('em', CellDfn('R0', style='em'))
Table.define_tmpl('pad',
    CellDfn('ALL', cls=Borders, border=(' ', 'm:0010')),
    CellDfn('C0', border='m:\n\n0\n'),
)
Table.define_tmpl('sep',
    CellDfn('R0', cls=Borders, border=('b:─-', 'c:\n\n──\n\n--', 'm:\n1\n\n')),
    CellDfn('R0C0', border=('c:\n\n├\n\n\n|\n')),
    CellDfn('R0C-1', border=('c:\n\n\n┤\n\n\n|')),
)
Table.define_tmpl(' ',
    CellDfn('ALL', cls=Borders, border=(' ','m:1010')),
    CellDfn('R-1', border=('m:1110')),
)
Table.define_tmpl('grid',
    CellDfn('ALL', cls=Borders, border=('-','c:┼┤┴┘++++','m:1010')),
    CellDfn('R0', border=('c:┬┐\n\n++\n\n')),
    CellDfn('C0', border=('c:├\n└\n+\n+\n')),
    CellDfn('C-1', border=('m:1011')),
    CellDfn('R-1', border=('m:1110')),
    CellDfn('R-1C-1', border=('m:1')),
    CellDfn('R0C0', border=('c:┌\n\n\n+\n\n\n')),
)
Table.define_tmpl('dbl',
    CellDfn('R-1', cls=Borders, border=('c:\n\n╧╝\n\n##', 'b:═#', 'm:\n1\n\n')),
    CellDfn('C0', cls=Borders, border=('c:╟╤╚╧####', 'l:║#', 'm:\n\n1\n')),
    CellDfn('C-1', cls=Borders, border=('c:\n╢╧╝\n###','r:║#','m:\n\n\n1')),
    CellDfn('R0', cls=Borders, border=('c:╤╗\n\n##\n\n', 't:═#', 'm:1\n\n\n')),
    CellDfn('R0C0', border=('c:╔\n\n\n#\n\n\n')),
)
Table.define_tmpl('kv',
    #CellDfn('ALL', cls=Borders, border=(' ', 'm:0010')),
    CellDfn('C0', border='m:\n\n0\n', style='1', just='>'),
    #CellDfn('C1', cls=Borders, border=('-', 'm:1'), width_rate=1)
)




from .rich import Rich
from .line import Line, Just, justify_v
from .span import Span


class Text(Rich):
    ''' This is a list of Lines
    '''
    def __init__(self, *args, **kwargs):
        self.__lines = []
        super().__init__(*args, **kwargs)


    @property
    def lines(self):
        if not self.rich_stream: return self.__lines
        self.__lines.append([self.__lines.pop()] if self.__lines else [])
        while self.rich_stream:
            el, ctx = self.rich_stream.pop(0)
            if el == '\v':
                self.__lines.append([])
                continue
            if isinstance(el, Line):
                lines = [Line(**ctx).insert(0,el)] if ctx else [el]
            else:
                lines = list(el.flatten()) if hasattr(el, 'flatten') else [Line().insert(0,el)]
                if ctx:
                    for l in lines: l.ctx(**ctx)
            if not lines: continue
            self.__lines[-1].append(lines[0])
            for line in lines[1:]:
                self.__lines.append([line])
        # Turn arrays of lines into a single Line object
        for i, line in enumerate(reversed(self.__lines)):
            if isinstance(line, Line): break
            self.__lines[-1-i] = line[0].ctx_parent(self) if len(line) == 1 else Line(*line, parent=self)
        return self.__lines


    def flatten(self, w=0, h=0, **kwargs):
        rows = []
        for line in self.lines:
            rows += list(line.flatten(w=w))
        maxw = rows[0].width if len(rows) == 1 else max(*[r.width for r in rows]) if rows and w == 0 else w
        for row in rows: row.justify(maxw)
        yield from justify_v(rows, h, Just(self['justify'],'^'), Line(parent=self).insert(0, ' '*maxw))


    def each_child(self):
        yield from self.lines


    def clone(self, **kwargs):
        s = self.__class__(**kwargs)
        s.__lines = [x.clone(parent=s, **x.ctx_local) for x in self.lines]
        return s
    

    def __str__(self):
        return '\\v'.join(map(str, self.lines))


    def __repr__(self):
        return self.__class__.__name__+f'({str(self)})'



import pytest
from print_ext.pretty import pretty
from .printer_test import _printer
from print_ext.table import Table
from .testutil import printer

class Beautiful():
    def __init__(self, **kwargs):
        self.vals = kwargs
    
    def pretty(self, **kwargs):
        t = Table(0,0)
        self.vals.update(kwargs)
        for k,v in self.vals.items():
            t(k,'\t',pretty(v),'\t')
        return t 


@pytest.mark.skip()
def test_pretty_list():
    o,p = _printer(color=True)
    l = pretty(['the',33,(99,1),{'a':39,'tens':[10,20,30,40]}])
    p(l)
    print(o.getvalue())
    assert(o.getvalue() == '')


def test_pretty_beautiful():
    o, p = printer()
    b = Beautiful(a=3, b=4)
    p.pretty(b, b=5, c=6)
    print(o.getvalue())
    assert(o.getvalue() == 'a 3\nb 5\nc 6\n')
